"""
Projects service
"""

import logging
from typing import List

from flask_sqlalchemy import BaseQuery

from portal.async_celery import tasks
from portal.database.models import Course, Project, ProjectConfig, Submission, SubmissionState, User
from portal.service import filters, general, permissions
from portal.service.errors import ForbiddenError
from portal.service.general import get_new_name

log = logging.getLogger(__name__)


def _set_project_data(project: Project, data: dict) -> Project:
    allowed = ['name', 'description', 'codename', 'assignment_url']
    return general.update_entity(project, data, allowed=allowed)


class ProjectService:
    def __init__(self, project: Project = None):
        self._project = project

    @property
    def project(self) -> Project:
        return self._project

    def copy_project(self, target: Course) -> Project:
        """Copies a project to the target course

        Args:
            target(Course): Course instance

        Returns(Project): Copied project
        """
        new_name = get_new_name(self.project, target)
        new_project = Project(target, codename=new_name)
        new_project.description = self.project.description
        new_project.name = self.project.name
        new_project.set_config(**vars(self.project.config))
        return new_project

    def can_create_submission(self, user: User) -> bool:
        """Checks if a user may create a new submission.

        A user is permitted to create a new submission if there are no submissions
        created or queued for him in the given project.

        Args:
            user(User): the user who wants to crete a new submission

        Returns: whether the user can create a new submission in the project

        """
        conflicting_states = [SubmissionState.CREATED,
                              SubmissionState.READY, SubmissionState.QUEUED]
        conflict = Submission.query.filter(Submission.user == user) \
            .filter(Submission.project == self.project) \
            .filter(Submission.state.in_(conflicting_states)) \
            .first()

        return conflict is None

    def delete_project(self):
        """Deletes a project
        """
        log.info(f"[DELETE] Project {self.project.id} ({self.project.codename}).")
        general.delete_entity(self.project)

    def update_project(self, data: dict) -> Project:
        """Updates a project

        Args:
            data(dict): Data dictionary

        Returns(Project): Updated project
        """
        log.info(f"[UPDATE] Project {self.project.id} ({self.project.codename}): {data}.")
        return _set_project_data(self.project, data)

    def create_project(self, course: Course, data: dict) -> Project:
        """Creates a new project

        Args:
            course(Course): Course instance
            data(dict): Project data

        Returns(Project): Project instance
        """
        new_project = Project(course=course)
        self._project = new_project
        _set_project_data(project=new_project, data=data)
        log.info(f"[CREATE] Project for course {course.id} ({course.codename}): "
                 f"{new_project}")
        return new_project

    def update_project_config(self, data: dict) -> ProjectConfig:
        """Updates project config

        Args:
            data(dict): Config data

        Returns(ProjectConfig): Project config instance
        """
        course = self.project.course
        self.project.set_config(**data)
        general.write_entity(self.project)
        log.info(f"[UPDATE] Configuration for project {self.project.id} in '"
                 f"course {course.id} to {data}.")
        return self.project.config

    def query_submissions(self, user: User) -> BaseQuery:
        """Queries submissions for the project and filters them by user.

        Args:
            user(User): User instance
        """
        course = self.project.course
        return Submission.query.filter(Submission.user_id == user.id) \
            .filter(Submission.project_id == self.project.id) \
            .join(Course, Course.id == self.project.course_id).filter(Course.id == course.id)

    def find_project_submissions(self, user_id: str = None) -> List[Submission]:
        """Finds all project submissions. If a user id is specified, returns only submissions created
        by that user.

        Args:
            user_id(str): User id (optional)

        Returns(List[Submission]): A list of all submissions in a project if no user_id is specified.
        Otherwise a list of submissions created in the project by the specified user.

        """
        submissions = self.project.submissions
        if user_id:
            user = general.find_user(user_id)
            submissions = self.query_submissions(user)
        return submissions

    def list_projects(self, course: Course) -> list:
        """List of all projects
        Args:
            course(Course): Course instance
            client: Client instance

        Returns(list): list of projects

        """
        perm_service = permissions.PermissionsService(course=course)
        if perm_service.check.client(['view_course_full']):
            return course.projects
        elif perm_service.check.client(['view_course_limited']):
            return filters.filter_projects_from_course(course=course, user=perm_service.client)
        raise ForbiddenError(perm_service.client)

    def update_project_test_files(self):
        """ Sends a request to Storage to update the project's test_files to the newest version.
        """
        tasks.update_project_test_files.delay(self.project.course.id, self.project.id)
