"""
Roles service
"""

import logging
from typing import List

from portal.database.models import Client, Course, Role, RolePermissions
from portal.service import permissions
from portal.service.errors import ForbiddenError, PortalAPIError
from portal.service.filters import filter_roles_from_course
from portal.service.general import delete_entity, find_client, get_new_name, update_entity, \
    write_entity

log = logging.getLogger(__name__)


def _set_role_data(role: Role, data: dict):
    return update_entity(role, data, allowed=['name', 'description', 'codename'])


class RoleService:
    def __init__(self, role: Role = None):
        self._role = role

    @property
    def role(self) -> Role:
        return self._role

    def copy_role(self, target: Course, with_clients: str) -> Role:
        """Copies role to the target course
        Args:
            target(Course): Target Course resource
            with_clients(str): With users flag

        Returns(Role): Copied role
        """
        source = self.role
        log.info(f"[COPY] Role: {self.role.id} to course {target.id} "
                 f"with config: {with_clients}")
        new_role = self.__copy_new_role(target)
        if with_clients == "with_users":
            for clients in source.clients:
                clients.roles.append(new_role)
        return new_role

    def __copy_new_role(self, target):
        source = self.role
        new_name = get_new_name(source, target)
        new_role = Role(target, codename=new_name, description=source.description, name=source.name)
        new_role.set_permissions(**vars(source.permissions))
        return new_role

    def delete_role(self):
        """Deletes role for the DB
        """
        log.info(f"[DELETE] Role: {self.role.id} ({self.role.name})")
        delete_entity(self.role)

    def update_role(self, data: dict) -> Role:
        """Updates role based on data
        Args:
            data(dict): Data for role update
        Returns(Role): Updated role
        """
        log.info(f"[UPDATE] Role {self.role.id} to {self.role}.")
        return _set_role_data(self.role, data)

    def create_role(self, course: Course, **data) -> Role:
        """Creates new role
        Args:
            course(Course): Course instance

        Keyword Args:
            name(str): Name of the role
            description(str): Description of the role
            codename(str): Codename of the role

        Returns(Role): New role instance
        """
        new_role = Role(course=course)
        self._role = new_role
        _set_role_data(role=new_role, data=data)
        log.info(f"[CREATE] Role for course {course.id}: {new_role}")
        return new_role

    def update_permissions(self, perm: dict) -> RolePermissions:
        """Updates role permissions
        Args:
            perm(dict): Permissions dictionary

        Returns(RolePermissions): Role Permissions instance
        """
        self.role.set_permissions(**perm)
        write_entity(self.role)
        log.info(f"[UPDATE] Permissions for role {self.role.id} "
                 f"in course {self.role.course.id}: {perm}.")
        return self.role.permissions

    def find_clients(self, ids: List[str]) -> List[Client]:
        """Finds all clients based on their ids

        Args:
            ids(List[str]): List of client ids

        Returns(List[Client]): List of clients
        """
        return [find_client(i) for i in ids]

    def update_clients_membership(self, data: dict) -> Role:
        """Updates client membership in the role
        Args:
            data(dict): Data provided to update role membership

        Returns(Role): Updated role
        """
        clients_old = set(self.role.clients)
        clients = self.__add_users_to_role(clients_old, data)
        clients = self.__remove_users_from_role(clients, clients_old, data)
        self.role.clients = list(clients)
        write_entity(self.role)
        log.info(f"[UPDATE] Clients in Role {self.role.id}: {data}.")
        return self.role

    def __remove_users_from_role(self, clients, clients_old, data):
        clients_remove = data.get('remove')
        if clients_remove:
            to_remove = self.find_clients(data['remove'])
            clients = clients_old.difference(to_remove)
        return clients

    def __add_users_to_role(self, clients_old, data):
        clients_add = data.get('add')
        clients = set()
        if clients_add:
            to_add = self.find_clients(data['add'])
            clients = clients_old.union(to_add)
        return clients

    def add_client(self, client: Client):
        """Adds single client to the role
        Args:
            client(Client): Client instance
        Returns:
        """
        if client not in self.role.clients:
            self.role.clients.append(client)
            write_entity(self.role)
            log.info(f"[ADD] Client {client.id} to role "
                     f"{self.role.id} in course {self.role.course.id}.")
        else:
            log.info(f"[ADD] Client {client.id} is already "
                     f"in role {self.role.id} in course {self.role.course.id}: no change.")

    def remove_client(self, client: Client) -> Role:
        """Removes single client from the role
        Args:
            role(Role): Role instance
            client(Client): Client instance
        Returns(Role): Updated role
        """
        course = self.role.course
        try:
            self.role.clients.remove(client)
            write_entity(self.role)
        except ValueError:
            raise PortalAPIError(400, message=f"Could not remove client {client.id} "
                                              f"from role {self.role.id} in course {course.id}: "
                                              f"role does not contain client.")
        log.info(f"[REMOVE] Client {client.id} from role "
                 f"{self.role.id} in course {course.id}.")
        return self.role

    def list_roles(self, course: Course) -> List[Role]:
        """List of all roles
        Args:
            course(Course): Course instance
            client(Client): Client instance

        Returns(list): List of roles

        """
        perm_service = permissions.PermissionsService(course=course)
        if perm_service.check.client(['view_course_full']):
            return course.roles
        elif perm_service.check.client(['view_course_limited']):
            return filter_roles_from_course(course=course, client=perm_service.client)
        raise ForbiddenError(perm_service.client)
