"""
Users service
"""

from typing import List

from portal import ldap_wrapper, logger
from portal.database.models import Course, Group, Project, Review, Role, Submission, User
from portal.service import emails, general, permissions
from portal.service.errors import IncorrectCredentialsError, PortalAPIError
from portal.service.general import find_course, find_user

log = logger.get_logger(__name__)


class UserLdapService:
    def __init__(self, service: 'UserService'):
        self._service = service

    @property
    def service(self):
        return self._service

    @property
    def ldap(self):
        return ldap_wrapper

    @property
    def is_enabled(self) -> bool:
        return self.ldap.is_enabled

    @property
    def user(self):
        return self.service.user

    def extract_uco(self) -> int:
        """Get uco from the ldap
        Returns(int): User's uco
        """
        username = self.user.username
        user_dict = self.ldap.search_dict(f"uid={username},ou=People")
        log.debug(f"[LDAP] User ({username}): {user_dict}")
        if not user_dict:
            return 0
        desc = user_dict[0]['description']
        uco = [u.decode('utf-8') for u in desc if u.decode('utf-8').startswith('UCO=')]
        if not uco:
            return 0
        return int(uco[0][4:])


class UserService:
    def __init__(self, user: User = None):
        self._user = user
        self._ldap = UserLdapService(self)

    @property
    def user(self) -> User:
        return self._user

    @property
    def ldap(self) -> UserLdapService:
        return self._ldap

    @property
    def emails(self):
        return emails

    @property
    def client(self):
        return permissions.PermissionsService().client_owner

    def _set_user_data(self, data: dict, full: bool = False) -> User:
        if (data.get("is_admin") is not None) and full:
            self.user.is_admin = data['is_admin']
        if self.ldap.is_enabled:
            data['uco'] = self.ldap.extract_uco()
        return general.update_entity(self.user, data, allowed=['name', 'email', 'uco'])

    def create_user(self, **data):
        """Creates a new user

            Keyword Args:
                username(str): User's name
                uco(str): User's Uco

            Returns(User): Created user instance
            """
        data['is_admin'] = data.get('is_admin', False)
        username = data['username']
        new_user = User(username=username)
        self._user = new_user
        self._set_user_data(data, True)

        self.emails.notify_user(new_user, 'users', 'created',
                                context=dict(username=username))
        log.info(f"[CREATE] User: {new_user}")
        return new_user

    def update_user(self, data: dict):
        """Updates a user

        Args:
            data(dict): User's data

        Returns(User): Updated user
        """
        self._set_user_data(data, self.client.is_admin)
        log.info(f"[UPDATE] User {self.user.id} to {self.user}.")
        return self.user

    def update_password(self, data: dict):
        log.info(f"[UPDATE] User password: {self.user.id} by {self.client.id}")
        self.__require_param(data, 'new_password')
        if self.client == self.user and not self.user.is_admin:
            self.__check_old_password(data)
        self.user.set_password(data['new_password'])
        general.write_entity(self.user)
        self.emails.notify_user(self.user, 'users', 'password_update',
                                context=dict(username=self.user.username))
        return self.user

    def __check_old_password(self, data):
        self.__require_param(data, 'old_password')
        old_password = data['old_password']
        if not self.user.verify_password(old_password):
            raise IncorrectCredentialsError()

    def __require_param(self, data, param):
        if not data.get('param'):
            raise PortalAPIError(400, f'missing param: "{param}" for password change')

    def delete_user(self):
        """Deletes a user
        """
        log.info(f"[DELETE] User({self.user.username}): {self.user.id} .")
        general.delete_entity(self.user)

    def find_users(self, ids: List[str]) -> List[User]:
        """Finds all users based on their ids

        Args:
            ids(List[str]): List of user ids

        Returns(List[User]): List of users
        """
        return [find_user(i) for i in ids]

    def find_users_filtered(self, course_id: str = None,
                            group_id: str = None) -> List[User]:
        """Find all users, optionally filtered by course and group
        Args:
            course_id(str): Course id (optional)
            group_id(str): Group id (optional)

        Returns(List[User]): Filtered list of  users
        """
        users = User.query
        if course_id:
            users = users.join(User.roles) \
                .join(Course).filter((Course.id == course_id) | (Course.codename == course_id))
        if group_id:
            course = find_course(course_id)
            users = users.join(User.groups).filter(
                ((Group.id == group_id) | (Group.name == group_id)) & (
                        course.id == Group.course_id)
            )

        return users.all()

    def find_groups_filtered(self, course_id: str = None) -> List[Group]:
        """Find all groups for the user, optionally filtered by course.

        Args:
            user(User): User instance
            course_id(str): Course id (optional)

        Returns(List[Group]): List of all groups
        """
        if course_id:
            return self.user.get_groups_in_course(find_course(course_id))
        return self.user.groups

    def find_roles_filtered(self, course_id: str = None) -> List[Role]:
        """Get all roles for the user, optionally filtered by course

        Args:
            user(User): User instance
            course_id(str): Course id (optional)

        Returns(List[Role]):
        """
        roles = self.user.roles
        if course_id:
            roles = self.user.get_roles_in_course(find_course(course_id))
        return roles

    def find_reviews(self) -> List[Review]:
        """Gets reviews the user has contributed to.

        Args:
            user(User): User instance

        Returns(List[Review]):
        """
        reviews = []
        for item in self.user.review_items:
            if item not in reviews:
                reviews.append(item.review)
        return list(reviews)

    def find_submissions_filtered(self, course_id: str = None,
                                  project_ids: List[str] = None) -> List[Submission]:
        """Get all submissions of a user, optionally filtered by course and project.

        Args:
            user(User):
            course_id(str): Course id (optional)
            project_ids(List[str]): List of project ids (optional)

        Returns(List[Submission]): submissions in the course and projects
        """
        submissions = Submission.query.filter_by(user=self.user)
        # user.submissions cannot be used as the base query
        if course_id:
            submissions = submissions.join(Project)
            if project_ids:
                course = find_course(course_id)
                submissions = submissions.filter(
                    ((Project.id.in_(project_ids)) | (Project.name.in_(project_ids))) &
                    (Project.course == course)
                )
            submissions = submissions.join(Course).filter(
                (Course.id == course_id) | (Course.codename == course_id))
        return submissions

    def find_all_projects_for_user(self):
        return self.user.get_all_projects()
