import json

from portal.database.models import Secret, Worker
from tests.rest import utils


def test_list(client):
    response = utils.make_request(client, '/workers', method='get')
    assert response.status_code == 200
    assert response.mimetype == 'application/json'

    workers = utils.extract_data(response)
    assert len(workers) == 2
    db_workers = Worker.query.all()
    utils.assert_worker_in(db_workers, workers[0])
    utils.assert_worker_in(db_workers, workers[1])


def test_create(client):
    request_dict = dict(
        name="new_executor",
        url="127.0.0.1/worker",
    )
    db_workers_number = len(Worker.query.all())

    request_json = json.dumps(request_dict)
    response = utils.make_request(client, '/workers',
                                  data=request_json,
                                  headers={"content-type": "application/json"},
                                  method='post')
    assert response.status_code == 201
    assert response.mimetype == 'application/json'

    resp_new_worker = utils.extract_data(response)
    db_workers = Worker.query.all()
    assert len(db_workers) == db_workers_number + 1
    utils.assert_worker_in(db_workers, resp_new_worker)


def test_read(client):
    w = Worker.query.filter_by(name="executor").first()
    response = utils.make_request(client, f"/workers/{w.id}", method='get')
    assert response.status_code == 200
    assert response.mimetype == 'application/json'

    worker = utils.extract_data(response)

    utils.assert_worker(w, worker)


def test_update(client):
    w = Worker.query.filter_by(name="executor").first()
    request_dict = dict(
        url="new url",
        portal_secret="such safe",
    )
    request_json = json.dumps(request_dict)
    response = utils.make_request(client, f'/workers/{w.id}', data=request_json,
                                  headers={"content-type": "application/json"}, method='put')

    assert response.status_code == 204
    assert response.mimetype == 'application/json'

    worker_after_update: Worker = Worker.query.filter_by(
        name="executor").first()
    assert worker_after_update.url == "new url"
    assert worker_after_update.portal_secret == "such safe"


def test_delete(client):
    worker = Worker.query.filter_by(name="executor").first()
    response = utils.make_request(client, f'/workers/{worker.id}',
                                  headers={"content-type": "application/json"},
                                  method='delete')

    assert response.status_code == 204
    assert response.mimetype == 'application/json'

    assert not Worker.query.filter_by(name="executor").first()


def test_create_secret(client):
    worker = Worker.query.filter_by(name="executor").first()
    request_dict = dict(
        name="new_secret",
    )
    response = utils.make_request(client, f'/workers/{worker.id}/secrets',
                                  headers={"content-type": "application/json"},
                                  data=json.dumps(request_dict),
                                  method='post')

    assert response.status_code == 201
    assert response.mimetype == 'application/json'
    parsed = response.json
    assert parsed['value'] is not None
    assert parsed['id'] is not None

    secrets = Secret.query.filter_by(client=worker)
    assert len(secrets.all()) == 2
    assert secrets.filter_by(name='new_secret').first()


def test_list_secret(client):
    worker = Worker.query.filter_by(name="executor").first()
    response = utils.make_request(client, f'/workers/{worker.id}/secrets',
                                  headers={"content-type": "application/json"},
                                  method='get')

    assert response.status_code == 200
    assert response.mimetype == 'application/json'
    secrets = utils.extract_data(response)
    assert len(secrets) == 1


def test_delete_secret(client):
    worker = Worker.query.filter_by(name="executor").first()
    secret = Secret.query.filter_by(client=worker).first()
    response = utils.make_request(client, f'/workers/{worker.id}/secrets/{secret.id}',
                                  headers={"content-type": "application/json"},
                                  method='delete')

    assert response.status_code == 204

    assert Secret.query.filter_by(client=worker).count() == 0


def test_read_secret(client):
    worker = Worker.query.filter_by(name="executor").first()
    secret = Secret.query.filter_by(client=worker).first()
    response = utils.make_request(client, f'/workers/{worker.id}/secrets/{secret.id}',
                                  headers={"content-type": "application/json"},
                                  method='get')

    assert response.status_code == 200
    assert response.mimetype == 'application/json'
    response_secret = utils.extract_data(response)
    assert response_secret['id'] == secret.id
    assert response_secret['name'] == secret.name

    assert Secret.query.filter_by(client=worker).count() == 1


def test_update_secret(client):
    worker = Worker.query.filter_by(name="executor").first()
    request_dict = dict(
        name="new_name",
    )
    secret = Secret.query.filter_by(client=worker).first()
    response = utils.make_request(client, f'/workers/{worker.id}/secrets/{secret.id}',
                                  headers={"content-type": "application/json"},
                                  data=json.dumps(request_dict),
                                  method='put')

    assert response.status_code == 204
    updated_secret = Secret.query.filter_by(client=worker).first()
    assert updated_secret.id == secret.id
    assert updated_secret.name == "new_name"

    assert Secret.query.filter_by(client=worker).count() == 1


